#include "funciones.hpp"

using namespace std;
using namespace cv;

int noise_filter_kernel[5][5] = {
	{ 2, 4, 5, 4, 2},
	{ 4, 9,12, 9, 4},
	{ 5,12,15,12, 5},
	{ 4, 9,12, 9, 4},
	{ 2, 4, 5, 4 ,2}
};

int kernel_x[3][4] = {
	{-1, 0, 1, 0},
	{-2, 0, 2, 0},
	{-1, 0, 1, 0}
};

int kernel_y[3][4] = {
	{-1,-2,-1, 0},
	{ 0, 0, 0, 0},
	{ 1, 2, 1, 0}
};

mutex mtx;

/////////////////////////////////////////////////////////////////////////////
/////////////////// Funciones de control de flujo del programa //////////////
/////////////////////////////////////////////////////////////////////////////

void process_key(parameters& params, Mat& canvas){

  bool actualizo_parametro = false;
  int c =  waitKey(10);
  switch(c){
  case 27:
  	params.stop();
  	actualizo_parametro = true;
    break;
  case 32:
  	params.pausarRetomar();
  	actualizo_parametro = true;
    break;
  case 43:
    params.relacion_votaciones_radio += 0.01;
    actualizo_parametro = true;
    break;
  case 45:
    params.restarRelacioVotacionRadio(0.01);
    actualizo_parametro = true;
    break;
  case 119:
  	params.umbral_bordes += 1;
  	actualizo_parametro = true;
    break;
  case 115:
  	params.restarUmbralBordes(1);
  	actualizo_parametro = true;
    break;
  case 100:
  	params.cantidad_radios += 1;
  	actualizo_parametro = true;
    break;
  case 97:
  	params.restarCantidadRadios(1);
  	actualizo_parametro = true;
    break;
  case 52:
  	params.restarMinRadio(1);
    actualizo_parametro = true;
    break;
  case 54:
    params.min_radio += 1;
  	actualizo_parametro = true;
    break;
  case 56:
    params.radio_step += 1;
    actualizo_parametro = true;
    break;
  case 50:
    params.restarRadioStep(1);
    actualizo_parametro = true;
    break;
  case 99:
  	params.restarAnchoLinea(1);
  	actualizo_parametro = true;
    break;
  case 118:
    params.sumarAnchoLinea(1);
    actualizo_parametro = true;
    break;
  case 107:
    params.restarCantidadThreads(1);
    actualizo_parametro = true;
    break;
  case 108:
    params.cantidad_threads += 1;
    actualizo_parametro = true;
    break;
  case 103:
    params.asmCpp();
  	actualizo_parametro = true;
    break;
  case 104:
    params.ajustar_hough = true;
    break;
  case 98:
    params.ajustar_bordes = true;
    break;
  case 112:
    params.dibujar_trayectoria = true;
    break;
  }

  if (actualizo_parametro) {

    color fondo;
    if (params.pausado) {
      fondo.r = 75;
      fondo.g = 90;
      fondo.b = 225;
    } else {
      fondo.r = 130;
      fondo.g = 200;
      fondo.b = 225;
    }
    pintar_uniforme(canvas, fondo, recorte(0,params.rows,params.cols,params.cols+700));
    escribir(canvas, 50, params.cols+10, params.comandosInStr());
    escribir(canvas, 50, params.cols+330, params.parametrosInStr());
    imshow("Orga 2 TP Final", canvas);
    waitKey(10);
  }
}

void ajustarHough(parameters& params, Mat& canvas, VideoCapture& capture){
  
  circulo circuloActual;
  Mat colorFrame, ByWFrame, flipedByWFrame;
  
  capture >> colorFrame;
  cvtColor(colorFrame, ByWFrame, COLOR_BGR2GRAY);
  cvtColor(colorFrame, flipedByWFrame, COLOR_BGR2GRAY);

  recorte zonaAProcesar;
  zonaAProcesar.i1 = 0;
  zonaAProcesar.j1 = 0;
  zonaAProcesar.i2 = ByWFrame.rows;
  zonaAProcesar.j2 = ByWFrame.cols;

  while(!params.terminado){

    process_key(params, canvas);

    if (!params.pausado) {
      capture >> colorFrame;
      cvtColor(colorFrame, ByWFrame, COLOR_BGR2GRAY);
      flip(ByWFrame, flipedByWFrame, 1);
      circuloActual = hough_circulos(&flipedByWFrame, zonaAProcesar, params);
      if (circuloActual.valido) {
        circle(flipedByWFrame, Point(circuloActual.x, circuloActual.y), circuloActual.radio, CV_RGB(0,0,0), -1, 8);
        circle(flipedByWFrame, Point(circuloActual.x, circuloActual.y), floor(circuloActual.radio*0.8), CV_RGB(255,255,255), -1, 8);
        circle(flipedByWFrame, Point(circuloActual.x, circuloActual.y), floor(circuloActual.radio*0.6), CV_RGB(0,0,0), -1, 8);
        circle(flipedByWFrame, Point(circuloActual.x, circuloActual.y), floor(circuloActual.radio*0.4), CV_RGB(255,255,255), -1, 8);
      }
      copiar_WyB_a_color(flipedByWFrame,canvas,recorte(0,params.rows,0,params.cols));
      imshow("Orga 2 TP Final", canvas);
    }
  }
  params.restart();

  ByWFrame.release();
  flipedByWFrame.release();
}

void ajustarBordes(parameters& params, Mat& canvas, VideoCapture& capture){
  
  list<borde>* bordesDeSrc = new list<borde>;  
  Mat colorFrame, ByWFrame;
  capture >> colorFrame;
  Mat flipedByWFrame(colorFrame.rows, colorFrame.cols, CV_8UC1);

  recorte zonaAProcesar;
  zonaAProcesar.i1 = 0;
  zonaAProcesar.j1 = 0;
  zonaAProcesar.i2 = colorFrame.rows;
  zonaAProcesar.j2 = colorFrame.cols;
  while(!params.terminado) {
    
    process_key(params, canvas);

    if (!params.pausado) {
      capture >> colorFrame;
      pintar_uniforme(canvas, color(0,0,0), recorte(0,params.rows,0,params.cols));
      cvtColor(colorFrame, ByWFrame, COLOR_BGR2GRAY);
      flip(ByWFrame, flipedByWFrame, 1);
      bordesDeSrc->clear();
      canny(&flipedByWFrame, zonaAProcesar, params, bordesDeSrc);
      dibujar_bordes(canvas, *bordesDeSrc);
      imshow("Orga 2 TP Final", canvas);
    }
  }
  params.restart();

  delete bordesDeSrc;
  ByWFrame.release();
  flipedByWFrame.release();
}

void pintarConHough(parameters& params, Mat& canvas, VideoCapture& capture){

  ///////////////////////////////////////////////////
  ///////////// DEBUGING PARAMETERS //////////////////
  ////////////////////////////////////////////////////

  double tiempo_total;
  clock_t t_total     = 0;
  clock_t t_init      = 0;
  clock_t t_fin       = 0;
  clock_t cant_ciclos = 0;

  ////////////////////////////////////////////////////
  //////////// FIN DEBUGING PARAMETERS ///////////////
  ////////////////////////////////////////////////////
  
  Mat colorFrame, ByWFrame, flipedByWFrame;
  circulo circuloActual;
  circulo circuloAnterior;
  int cantidadSalteados = 0;

  capture >> colorFrame;
  recorte zonaAProcesar;
  zonaAProcesar.i1 = 0;
  zonaAProcesar.j1 = 0;
  zonaAProcesar.i2 = colorFrame.rows;
  zonaAProcesar.j2 = colorFrame.cols;

  pintar_uniforme(canvas, color(200,200,200), recorte(0,params.rows,0,params.cols));


  while(!params.terminado){
    ++cant_ciclos;
    t_init = clock();

    process_key(params, canvas);

    if (!params.pausado) {
      capture >> colorFrame;
      cvtColor(colorFrame, ByWFrame, COLOR_BGR2GRAY);
      flip(ByWFrame,flipedByWFrame,1);
      circuloActual = hough_circulos(&flipedByWFrame, zonaAProcesar, params);
      if (circuloActual.valido) {
        if (dist(circuloActual,circuloAnterior) < params.distancia_maxima) {
          circle(canvas, Point(circuloActual.x, circuloActual.y), params.ancho_linea - 2, CV_RGB(0,0,0), -1, 8);
          line(canvas, Point(circuloActual.x, circuloActual.y), cvPoint(circuloAnterior.x, circuloAnterior.y), CV_RGB(0,0,0), params.ancho_linea);
          circuloAnterior = circuloActual;
          imshow("Orga 2 TP Final", canvas);
        } else if (cantidadSalteados >= params.cantidad_lejanos_maxima) {
          circle(canvas, Point(circuloActual.x, circuloActual.y), params.ancho_linea - 2, CV_RGB(0,0,0), -1, 8);
          circuloAnterior = circuloActual;
          imshow("Orga 2 TP Final", canvas);
          cantidadSalteados = 0;
        } else {
          ++cantidadSalteados;
        }
      }
    }
    t_fin = clock();
    t_total += (t_fin - t_init);
  }

  t_total = t_total / cant_ciclos;
  tiempo_total = (double)t_total / CLOCKS_PER_SEC;

  ByWFrame.release();
  flipedByWFrame.release();

  params.restart();

  ////////////////////////////////////////////////////
  //////////// SECCION PARA DEBUGING /////////////////
  ////////////////////////////////////////////////////

  cout << "ciclos por segundos: " << tiempo_total << endl;
  cout << "anda: " << tiempo_total / 0.0416666666666667 << " veces mas lento de lo que se necesita para ir a 24 FPS" << endl;

  ////////////////////////////////////////////////////
  ////////// FIN SECCION PARA DEBUGING ///////////////
  ////////////////////////////////////////////////////
}

/////////////////////////////////////////////////////////////////////////////
//////////////////////////// Funciones de hougs /////////////////////////////
/////////////////////////////////////////////////////////////////////////////


uchar convolucion_suavizado_cpp(uchar* src, int y, int x, int alto, int ancho, int step) {

  // declaro variables
  int contador  = 0;

  for (int i = 0; i < 5; ++i){
    for (int j = 0; j < 5; ++j){
      contador += ((int) src[(y-2+i)*step + (x-2+j)]) * noise_filter_kernel[i][j];
    }
  }

  return (uchar) ((int) contador/159);
}

int convolucion_gradiente_cpp(uchar* src, int y, int x, int ancho, int kernel[3][4]) {

	// declaro variables
	int sumatoria = 0;

	for (int i = 0; i < 3; i++){
		for (int j = 0; j < 3; j++){
			sumatoria += ((int) src[(y-1+i)*ancho+(x+j)]) * kernel[i][j];
		}
	}

	return sumatoria;
}

void suavizar_cpp(uchar* imgSuavizada, recorte zonaAProcesar, int ancho, int alto, int stepSrc, uchar* dataSrc) {
  for (int i = zonaAProcesar.i1; i < zonaAProcesar.i2; ++i){
    for (int j = zonaAProcesar.j1; j < zonaAProcesar.j2; ++j){
      imgSuavizada[(i*stepSrc + j)] = convolucion_suavizado_cpp(dataSrc, i, j, alto, ancho, stepSrc);
    }
  }
}

void suavizar_asm(uchar* imgSuavizada, recorte zonaAProcesar, int ancho, int alto, int stepSrc, uchar* dataSrc) { // el parametro ancho está para que la función tome los mismos parámetros que la versión cpp
  convolucion_suavizado_asm(dataSrc, &zonaAProcesar, imgSuavizada, stepSrc, alto);
}

void obtener_bordes_cpp(recorte zonaAProcesar, uchar* imgSuavizada, int ancho, int alto, list<borde>* bordes, int umbral_bordes) {

  double gx, gy, mod_gradiente;
  borde mi_borde;

  for (int i = zonaAProcesar.i1; i < zonaAProcesar.i2; ++i) {
    for(int j = zonaAProcesar.j1; j < zonaAProcesar.j2; ++j) {
      gx = convolucion_gradiente_cpp(imgSuavizada,i,j,ancho,kernel_x);
      gy = convolucion_gradiente_cpp(imgSuavizada,i,j,ancho,kernel_y);

      mod_gradiente = sqrt(gx*gx + gy*gy);
      if ((int)mod_gradiente >= umbral_bordes){

        mi_borde.posicion.first = i;
        mi_borde.posicion.second = j;
        mi_borde.gradiente = mod_gradiente;
        mi_borde.theta = atan2(gy,gx);
        mtx.lock();
        bordes->push_back(mi_borde);
        mtx.unlock();
      }
    }
  }
}

void obtener_bordes_asm(recorte zonaAProcesar, uchar* imgSuavizada, int ancho, int alto, list<borde>* bordes, int umbral_bordes) {

  double gx, gy, mod_gradiente;
  int buffer_x[4];
  int buffer_y[4];
  borde mi_borde;

  int step = ancho;
  for (int i = zonaAProcesar.i1; i < zonaAProcesar.i2; ++i) {    
    for (int j = zonaAProcesar.j1; j < zonaAProcesar.j2; j = j+4){ // TODO(lao): no estoy seguro de esto.    
      convolucion_gradiente_asm(imgSuavizada, i, j, step, &(kernel_x[0][0]), &buffer_x[0]);
      convolucion_gradiente_asm(imgSuavizada, i, j, step, &(kernel_y[0][0]), &buffer_y[0]);

      // Ahora proceso los 4 valores uno a uno
      for (int conv_it = 0; conv_it < 4; ++conv_it){
        gx = (double) buffer_x[conv_it];
        gy = (double) buffer_y[conv_it];

        mod_gradiente = sqrt(gx*gx + gy*gy);

        if ((int)mod_gradiente >= umbral_bordes){

          mi_borde.posicion.first = i;
          mi_borde.posicion.second = j+conv_it;
          mi_borde.gradiente = mod_gradiente;
          mi_borde.theta = atan2(gy,gx);

          mtx.lock();
          bordes->push_back(mi_borde);
          mtx.unlock();
        }
      }
    }
    
  }
}

void canny(Mat* src, const recorte &zonaAProcesar, parameters params, list<borde>* bordes) {

	int alto = src->rows;
	int ancho = src->cols;
	int step = src->step1();
	uchar* data = src->data;

	// nueva matris en donde alojar la imagen suavizada
	uchar* imgSuavizada = new uchar[alto * step];

	// aplicar filtro de ruido con threads
	recorte zonaActual;
	int altoZona, anchoZona;

	vector<thread> threadSuavizar;
	altoZona = floor((zonaAProcesar.alto()) / params.cantidad_threads);
	anchoZona = floor((zonaAProcesar.ancho()) / params.cantidad_threads);
	for (int i = 0; i < params.cantidad_threads; ++i){
		for (int j = 0; j < params.cantidad_threads; ++j){
			zonaActual.i1 = max(2, zonaAProcesar.i1 + i*altoZona);
			zonaActual.i2 = min(alto-2, zonaAProcesar.i1 + i*altoZona + altoZona);
			zonaActual.j1 = max(2, zonaAProcesar.j1 + j*anchoZona);
			zonaActual.j2 = min(ancho-2, zonaAProcesar.j1 + j*anchoZona + anchoZona);
      if (params.usar_asm)
        threadSuavizar.push_back(thread(suavizar_asm, imgSuavizada, zonaActual, ancho, alto, step, data));
      else
        threadSuavizar.push_back(thread(suavizar_cpp, imgSuavizada, zonaActual, ancho, alto, step, data));
		}
	}

	for (vector<thread>::iterator it = threadSuavizar.begin(); it != threadSuavizar.end(); ++it){
		it->join();
	}

	// ya tengo la imagen suavizada en imgSuavizada
	vector<thread> threadBordes;

  for (int i = 0; i < params.cantidad_threads; ++i){
    for (int j = 0; j < params.cantidad_threads; ++j){
      zonaActual.i1 = max(3, zonaAProcesar.i1 + i*altoZona);
      zonaActual.i2 = min(alto-3, zonaAProcesar.i1 + i*altoZona + altoZona);
      zonaActual.j1 = max(3, zonaAProcesar.j1 + j*anchoZona);
      zonaActual.j2 = min(ancho-18, zonaAProcesar.j1 + j*anchoZona + anchoZona);
      if (params.usar_asm)
        threadBordes.push_back(thread(obtener_bordes_asm, zonaActual, imgSuavizada, ancho, alto, bordes, params.umbral_bordes));
      else
        threadBordes.push_back(thread(obtener_bordes_cpp, zonaActual, imgSuavizada, ancho, alto, bordes, params.umbral_bordes));
    }
  }

  for (vector<thread>::iterator it = threadBordes.begin(); it != threadBordes.end(); ++it){
    it->join();
  }

//  mostrar_img(imgSuavizada, alto, step);
  delete imgSuavizada;
}

circulo hough_circulos(Mat* src, recorte &zonaAProcesar, const parameters& params){

	list<borde>* bordesDeSrc = new list<borde>;
	uint maximaVotacionCirculo = 0;
	int alto = src->rows;
	int ancho = src->cols;
	int area = ancho * alto;
	circulo* circuloEncontradoPorRadio = new circulo[params.cantidad_radios];
	uint* maxVotaciones = new uint[params.cantidad_radios];

	int* iMax_Lst = new int[params.cantidad_radios];
	int* jMax_Lst = new int[params.cantidad_radios];
	int* radios_Lst = new int[params.cantidad_radios];

  canny(src,zonaAProcesar, params, bordesDeSrc);

	vector<thread> threadsVotaciones;

  for (int r = 0; r < params.cantidad_radios; ++r){
    maxVotaciones[r] = 0;
    int _radio = params.min_radio + r*params.radio_step;
    radios_Lst[r] = _radio;
    threadsVotaciones.push_back(thread(votacion, bordesDeSrc, circuloEncontradoPorRadio, maxVotaciones, iMax_Lst, jMax_Lst, r, ancho, alto, area, _radio));

  }

  for (int r = 0; r < params.cantidad_radios; ++r){
    threadsVotaciones[r].join();
  }

	circulo circuloEncontrado;
	circuloEncontrado.valido = false;

	int maximo_indice = 0;

	for (int r = 0; r < params.cantidad_radios; ++r){
		if (maxVotaciones[r] > maximaVotacionCirculo){
			circuloEncontrado = circuloEncontradoPorRadio[r];
			maximaVotacionCirculo = maxVotaciones[r];
			maximo_indice = r;
		}
	}

  circuloEncontrado.cantidad_votos = maximaVotacionCirculo;

	int iMax = iMax_Lst[maximo_indice];
	int jMax = jMax_Lst[maximo_indice];
	int radio = radios_Lst[maximo_indice];
	//////////////////////////////////////////////
	//////// SECCION REDEFINIR ZONA //////////////
	//////////////////////////////////////////////

	if ( (double)maximaVotacionCirculo > (double)(params.relacion_votaciones_radio*(double)circuloEncontrado.radio)){

		zonaAProcesar.i1 = ((iMax - 2*radio) < 0) 		? 0 	: iMax - 2*radio;
		zonaAProcesar.i2 = ((iMax + 2*radio) > alto) 	? alto 	: iMax + 2*radio;
		zonaAProcesar.j1 = ((jMax - 2*radio) < 0) 		? 0 	: jMax - 2*radio;
		zonaAProcesar.j2 = ((jMax + 2*radio) > ancho) 	? ancho : jMax + 2*radio;

	}else{
		
		circuloEncontrado.valido = false;
		zonaAProcesar.i1 = 0;
		zonaAProcesar.j1 = 0;
		zonaAProcesar.i2 = alto;
		zonaAProcesar.j2 = ancho;
	}

	//////////////////////////////////////////////
	////// FIN SECCION REDEFINIR ZONA ////////////
	//////////////////////////////////////////////

	//////////////////////////////////////////////
	//////// SECCION LIBERAR MEMORIA  ////////////
	//////////////////////////////////////////////

	delete circuloEncontradoPorRadio;
	delete maxVotaciones;
	delete iMax_Lst;
	delete jMax_Lst;
	delete radios_Lst;
	delete bordesDeSrc;

	return circuloEncontrado;
}

void votacion(list<borde>* bordesDeSrc, circulo* circuloEncontradoPorRadio, uint* maxVotaciones, int* iMax_Lst, int* jMax_Lst, int r, int ancho, int alto, int area, int radio){

	// seteo las votaciones en 0
	circulo circuloEncontrado;
	circuloEncontrado.valido = false;
	int iCentro,jCentro;
	double thetaBorde;
	bool estaDentroDeLaImagen;
	uint* votaciones = new uint[area];
	for (int i = 0; i < area; ++i){
		votaciones[i] = 0;
	}
	
	////////////////////////////////////////////////////
	////////////// SECCION VOTACIONES //////////////////
	////////////////////////////////////////////////////

	// para cada punto de borde voto los 2 posibles puntos del centro obtenido con el theta de canny
	for (list<borde>::const_iterator bordeActual = bordesDeSrc->cbegin(); bordeActual != bordesDeSrc->cend(); ++bordeActual){

		thetaBorde = bordeActual->theta;

		// Primera posibilidad sumando sin y cos
		iCentro = bordeActual->posicion.first + sin(thetaBorde)*radio;
		jCentro = bordeActual->posicion.second + cos(thetaBorde)*radio;
		estaDentroDeLaImagen = (0 <= jCentro) and (jCentro < ancho) and (0 <= iCentro) and (iCentro < alto);

		if(estaDentroDeLaImagen) ++votaciones[iCentro*ancho + jCentro];

		// Segunda posibilidad restando sin y cos
		iCentro = bordeActual->posicion.first - sin(thetaBorde)*radio;
		jCentro = bordeActual->posicion.second - cos(thetaBorde)*radio;
		estaDentroDeLaImagen = (0 <= jCentro) and (jCentro < ancho) and (0 <= iCentro) and (iCentro < alto);

		if(estaDentroDeLaImagen) ++votaciones[iCentro*ancho + jCentro];
	}

	////////////////////////////////////////////////////
	///////////// SECCION ELEGIR PUNTO /////////////////
	////////////////////////////////////////////////////

	// cambiar para usar los bordes o la zona recortada pero no toda las votaciones
	int iMax = 0;
	int jMax = 0;
	for (int i = 0; i < alto; ++i){
		for (int j = 0; j < ancho; ++j){

			if (votaciones[iMax*ancho + jMax] < votaciones[i*ancho + j]) {

				iMax = i;
				jMax = j;
			}
		}
	}

	iMax_Lst[r] = iMax;
	jMax_Lst[r] = jMax;

	uint votacionesCirculoActual = votaciones[iMax*ancho + jMax];

	if (votaciones[iMax*ancho + jMax] > 0){
		circuloEncontrado.x = jMax; // Esto es para espejar la imagen
		circuloEncontrado.y = iMax;
		circuloEncontrado.valido = true;
		circuloEncontrado.radio = radio;
	}

	circuloEncontradoPorRadio[r] = circuloEncontrado;
	maxVotaciones[r] = votacionesCirculoActual;
	delete votaciones;
}
/////////////////////////////////////////////////////////////////////////////
//////////////////////////// Funciones auxiliares ///////////////////////////
/////////////////////////////////////////////////////////////////////////////

void pintar_uniforme(Mat& src, color colorASetear, recorte zonaAProcesar) {

  if (zonaAProcesar.i1 == -1) {
    zonaAProcesar.i1 = 0;
    zonaAProcesar.i2 = src.rows;
    zonaAProcesar.j1 = 0;
    zonaAProcesar.j2 = src.cols;  
  }

  int step = src.step1();
  uchar* imagen = src.data;

  for (int i = zonaAProcesar.i1; i < zonaAProcesar.i2; ++i){
    for(int j = zonaAProcesar.j1; j < zonaAProcesar.j2; ++j){
      imagen[i*step+3*j] = colorASetear.r;
      imagen[i*step+3*j+1] = colorASetear.g;
      imagen[i*step+3*j+2] = colorASetear.b;
    }
  }
}

//deprecated
void gray_setear_en_negro(Mat& src){

	int step = src.step1();

	for (int i = 0; i < src.rows; ++i){
		for(int j = 0; j < src.cols; j++){
			src.data[i*step+j] = 0;
		}
	}
}

void dibujar_bordes(Mat& src, const list<borde>& bordes){

	int step = src.step1();
	uchar* imagen = src.data;

  for (list<borde>::const_iterator it = bordes.cbegin(); it != bordes.cend(); ++it){

    imagen[it->posicion.first*step+3*it->posicion.second] = 255;
		imagen[it->posicion.first*step+3*it->posicion.second+1] = 255;
		imagen[it->posicion.first*step+3*it->posicion.second+2] = 255;
	}
}

double dist(circulo c1, circulo c2) {
  
  return std::sqrt(std::pow(c1.x-c2.x,2) + std::pow(c1.y-c2.y,2));
}

void copiar_WyB_a_color(Mat& src, Mat& dst, recorte zonaACopiar) {

  int srcStep = src.step1();
  int dstStep = dst.step1();

  for (int i = zonaACopiar.i1; i < zonaACopiar.i1 + zonaACopiar.i2; ++i){
    for(int j = zonaACopiar.j1; j < zonaACopiar.j1 + zonaACopiar.j2; j++){

      dst.data[i*dstStep+3*j] = src.data[i*srcStep+j];
      dst.data[i*dstStep+3*j+1] = src.data[i*srcStep+j];
      dst.data[i*dstStep+3*j+2] = src.data[i*srcStep+j];
    }
  }
}

void escribir(Mat& canvas, int i, int j, const vector<pair<String,color>>& lineas) {
  
  for (uint k = 0; k < lineas.size(); ++k) {
    color color = lineas[k].second;
    putText(canvas, lineas[k].first, Point(j, i+k*20), FONT_HERSHEY_SIMPLEX, 0.5, Scalar(color.b,color.g,color.r));
  }
}

void mostrar_img(uchar *img, int alto, int ancho) {
  Mat imagen(alto,ancho,CV_8UC1);

  for (int i = 0; i < alto; ++i) {
    for (int j = 0; j < ancho; ++j) {
      imagen.data[i*imagen.step1()+j] = img[i*ancho+j];
    }
  }

  imshow("debugger", imagen);
  imagen.release();
}