#ifndef ___FUNCIONES_H___
#define ___FUNCIONES_H___
#include <opencv/cv.h>
#include <opencv/highgui.h>
#include <iostream>
#include <fstream>
#include <stdio.h>
#include <math.h>
#include <list>
#include <utility>
#include <vector>
#include <time.h>
#include <thread>
#include <mutex>
#include <exception>

using namespace std;
using namespace cv;

struct recorte{

  int i1,i2,j1,j2;

  int alto() const {
    return i2-i1;
  };

  int ancho() const {
    return j2-j1;
  };

  void show()  const {
    cout << i1 << " " << i2 << " " << j1 << " " << j2 << endl;
  }

  recorte()
    : i1(-1), i2(-1), j1(-1), j2(-1) {};

  recorte(int oi1, int oi2, int oj1, int oj2)
    : i1(oi1), i2(oi2), j1(oj1), j2(oj2) {};
};

struct circulo{

  int x,y,radio,cantidad_votos;
  bool valido;
  circulo(): x(0), y(0), radio(0), cantidad_votos(0), valido(true) {}

  circulo(int x_other, int y_other, int radio_other)
  : x(x_other), y(y_other), radio(radio_other), cantidad_votos(0), valido(true){}

  struct circulo& operator=(const circulo other){
    this->x = other.x;
    this->y = other.y;
    this->radio = other.radio;
    this->cantidad_votos = other.cantidad_votos;
    this->valido = other.valido;

    return *this;
  }

  void show(){
    cout << "(" << this->x << "," << this->y <<") radio:" << this->radio << endl << ((valido) ? "circulo valido" : "circulo invalido") << "cantidad de votos: " << to_string(this->cantidad_votos) << endl;
  }
};

struct borde {

  // posicion = <i,j>
  std::pair<unsigned int, unsigned int> posicion;
  double gradiente;
  double theta;

  borde(){

    posicion.first 	= 0;
    posicion.second = 0;
  }
};

struct color{

  int r;
  int g;
  int b;

  color(int r_other, int g_other, int b_other)
    : r(r_other), g(g_other), b(b_other){}
  color()
    : r(0), g(0), b(0){}
  int valor(int i) {
    int result = -1;
    if(i == 0) result = this->b;
    else if(i == 1) result = this->g;
    else if(i == 2) result = this->r;

    return result;
  }
};

struct parameters{
  bool usar_asm;
  bool terminado;
  bool pausado;
  double relacion_votaciones_radio;
  int umbral_bordes;
  int min_radio;
  int cantidad_radios;
  int radio_step;
  int ancho_linea;
  int distancia_maxima;
  int cantidad_lejanos_maxima;
  int rows;
  int cols;
  int cantidad_threads;
  bool ajustar_bordes;
  bool ajustar_hough;
  bool dibujar_trayectoria;

  parameters(double rvr, int bt, int mr, int cr, int rs, int al, int dm, int clm, int ct) : 
    usar_asm(false),
    terminado(false), 
    pausado(false), 
    relacion_votaciones_radio(rvr), 
    umbral_bordes(bt), 
    min_radio(mr), 
    cantidad_radios(cr), 
    radio_step(rs), 
    ancho_linea(al), 
    distancia_maxima(dm), 
    cantidad_lejanos_maxima(clm),
    cantidad_threads(ct),
    ajustar_bordes(false),
    ajustar_hough(false),
    dibujar_trayectoria(false) {}

  void setNuevosValores(double rvr, int bt, int mr, int cr, int rs, int al, int dm, int clm, int ct, bool ua) {
    relacion_votaciones_radio = rvr;
    umbral_bordes = bt;
    min_radio = mr;
    cantidad_radios = cr;
    radio_step = rs;
    ancho_linea = al;
    distancia_maxima = dm;
    cantidad_lejanos_maxima = clm;
    cantidad_threads = ct;
    usar_asm = ua;
  }

  void pausarRetomar() {pausado = !pausado;}

  void stop() {terminado = true;}

  void asmCpp() {usar_asm = !usar_asm;}

  void restart() {
    terminado = false;
    ajustar_hough = false;
    ajustar_bordes = false;
    dibujar_trayectoria = false;
  }

  vector<pair<String,color>> comandosInStr() {
    vector<pair<String,color>> res;
    color negro(0,0,0);
    res.push_back(make_pair("COMANDOS:",negro));
    res.push_back(make_pair("Pausar / Retomar: espacio",negro));
    res.push_back(make_pair("C++ / ASM: g",negro));
    res.push_back(make_pair("Ajustar bordes: b",negro));
    res.push_back(make_pair("Ajustar Hough: h",negro));
    res.push_back(make_pair("Dibujar trayectoria: p",negro));
    res.push_back(make_pair("Terminar tarea/programa: esc",negro));
    res.push_back(make_pair("Relacion votaciones radio: + / -",negro));
    res.push_back(make_pair("Umbral bordes: w / s",negro));
    res.push_back(make_pair("Radio minimo: 6 / 4",negro));
    res.push_back(make_pair("Radio step: 8 / 2",negro));
    res.push_back(make_pair("Cantidad de radios: d / a",negro));
    res.push_back(make_pair("Ancho de linea: v / c",negro));
    res.push_back(make_pair("Cantidad de threads: l / k",negro));
    return res;
  }

  vector<pair<String,color>> parametrosInStr() {
    vector<pair<String,color>> res;
    color negro(0,0,0);
    color rojo(112,13,0);
    color verde(9,150,51);
    color violeta(50,35,90);
    color marron(140,95,46);
    res.push_back(make_pair("PARAMETROS:",negro));
    res.push_back((pausado ? make_pair("Pausado", rojo): make_pair("Corriendo", verde)));
    res.push_back((usar_asm ? make_pair("ASM", violeta): make_pair("C++", marron)));
    res.push_back(make_pair("Relacion votaciones radio: " + to_string(relacion_votaciones_radio), negro));
    res.push_back(make_pair("Umbral bordes: " + to_string(umbral_bordes), negro));
    res.push_back(make_pair("Radio minimo: " + to_string(min_radio), negro));
    res.push_back(make_pair("Radio step: " + to_string(radio_step), negro));
    res.push_back(make_pair("Cantidad de radios: " + to_string(cantidad_radios), negro));
    res.push_back(make_pair("Ancho linea: " + to_string(ancho_linea), negro));
    res.push_back(make_pair("Cantidad de threads: " + to_string(cantidad_threads*cantidad_threads), negro));
    return res;
  }

  void mostrarComandos() {
    cout << "Comandos:" << endl;

    cout << "\t terminar rograma: esc" << endl;
    cout << "\t mostrar comandos: c" << endl;
    cout << "\t pausar: espacio" << endl;
    cout << "\t retomar: espacio" << endl;
    cout << "\t mostrar parametros actuales: enter" << endl;
    cout << "\t relacion votaciones radio: + / -" << endl;
    cout << "\t border threshold: w / s" << endl;
    cout << "\t radio minimo: 6 / 4" << endl;
    cout << "\t cantidad de radios: d / a" << endl;
    cout << "\t radio step: 8 / 2" << endl;
    cout << "\t ancho de linea: v / c" << endl;
    cout << "\t Cantidad de threads: k / l" << endl;
  }

  void restarRelacioVotacionRadio(double a) {
    if (this->relacion_votaciones_radio - a >= 0)
      this->relacion_votaciones_radio -= a;
  }

  void restarUmbralBordes(int a) {
    if (this->umbral_bordes - a >= 0)
      this->umbral_bordes -= a;
  }

  void restarMinRadio(int a) {
    if(this->min_radio - a >= 0) 
      this->min_radio -= a;
  }

  void restarCantidadRadios(int a) {
    if (this->cantidad_radios - a >= 1) 
      this->cantidad_radios -= a;
  }

  void restarRadioStep(int a) {
    if (this->radio_step - a >= 1)
      this->radio_step -= a;
  }

  void restarAnchoLinea(int a) {
    if (this->ancho_linea - a >= 1)
      this->ancho_linea -= a;
  }

  void sumarAnchoLinea(int a) {
    if (this->ancho_linea + a <= 50)
      this->ancho_linea += a;
  }

  void restarDistanciaMaxima(int a){
    if (this->distancia_maxima - a >= 1) {
      this->distancia_maxima -= a;
    }
  }

  void restarCantidadLejanosMaxima(int a) {
    if (this->cantidad_lejanos_maxima - a >= 1)
      this->cantidad_lejanos_maxima -= a;
  }

  void restarCantidadThreads(int a) {
    if (this->cantidad_threads - a >= 1)
      this->cantidad_threads -= a;
  }
};

/////////////////////////////////////////////////////////////////////////////
/////////////////// Funciones de control de flujo del programa //////////////
/////////////////////////////////////////////////////////////////////////////

void process_key(parameters& params, Mat& canvas);

void ajustarHough(parameters& params, Mat& canvas, VideoCapture& capture);

void ajustarBordes(parameters& params, Mat& canvas, VideoCapture& capture);

void pintarConHough(parameters& params, Mat& canvas, VideoCapture& capture);

/////////////////////////////////////////////////////////////////////////////
//////////////////////////// Funciones de hougs /////////////////////////////
/////////////////////////////////////////////////////////////////////////////

uchar convolucion_suavizado_cpp(uchar* src, int i, int j, int alto, int ancho, int step_src);

int convolucion_gradiente_cpp(uchar* src, int i, int j, int step, int kernel[3][4]);

extern "C" void convolucion_suavizado_asm(unsigned char* src, recorte* zonaAProcesar, uchar* imgSuavizada, int stepSrc, int alto);

extern "C" void convolucion_gradiente_asm(unsigned char*, int, int, int, int*, int*);

void obtener_bordes_cpp(recorte zonaAProcesar, uchar* imgSuavizada, int ancho, int alto, list<borde>* bordes, int umbral_bordes);

void obtener_bordes_asm(recorte zonaAProcesar, uchar* imgSuavizada, int ancho, int alto, list<borde>* bordes, int umbral_bordes);

void suavizar_cpp(uchar* imgSuavizada, recorte zonaAProcesar, int ancho, int alto, int stepSrc, uchar* dataSrc);

void canny(Mat* src, const recorte &zonaAProcesar, parameters params, list<borde>* bordes);

void votacion(list<borde>* bordesDeSrc, circulo* circuloEncontradoPorRadio, uint* maxVotaciones, int* iMax_Lst, int* jMax_Lst, int r, int ancho, int alto, int area, int radio);

circulo hough_circulos(Mat* src, recorte &zonaAProcesar, const parameters& params);

/////////////////////////////////////////////////////////////////////////////
//////////////////////////// Funciones auxiliares ///////////////////////////
/////////////////////////////////////////////////////////////////////////////

void pintar_uniforme(Mat& src, color colorASetear, recorte zonaAProcesar);

void gray_setear_en_negro(Mat& src);

void dibujar_bordes(Mat& src, const list<borde>& bordes);

double dist(circulo c1, circulo c2);

void copiar_WyB_a_color(Mat& src, Mat& dst, recorte zonaACopiar);

void escribir(Mat& canvas, int i, int j, const vector<pair<String,color>>& lineas);

void mostrar_img(uchar *img, int alto, int ancho);

#endif
